package be.iminds.ilabt.jfed.highlevel.jobs.states;

import be.iminds.ilabt.jfed.experiment.Experiment;
import be.iminds.ilabt.jfed.experiment.tasks.ExperimentTaskStatus;
import be.iminds.ilabt.jfed.highlevel.jobs.*;
import be.iminds.ilabt.jfed.highlevel.util.ExternalFileUtil;
import be.iminds.ilabt.jfed.highlevel.util.StringMemorySourceFile;
import be.iminds.ilabt.jfed.lowlevel.connection.JFedException;
import be.iminds.ilabt.jfed.rspec.model.DistributeSshKeypair;
import be.iminds.ilabt.jfed.rspec.model.ModelRspec;
import be.iminds.ilabt.jfed.rspec.model.RspecNode;
import be.iminds.ilabt.jfed.rspec.rspec_source.RequestRspecSource;
import be.iminds.ilabt.jfed.util.library.KeyUtil;
import net.schmizz.sshj.SSHClient;
import net.schmizz.sshj.connection.channel.direct.Session;
import net.schmizz.sshj.userauth.UserAuthException;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;

import javax.annotation.Nonnull;
import java.io.BufferedReader;
import java.io.IOException;
import java.io.InputStreamReader;
import java.security.KeyPair;
import java.util.Collections;
import java.util.List;
import java.util.stream.Collectors;

/**
 * User: twalcari
 * Date: 5/26/2016
 * Time: 16:22
 */
public class DistributeSshKeysState extends SlicedState<DistributeSshKeysState.DistributeSshKeysStateSlice> {
    private static final Logger LOG = LoggerFactory.getLogger(DistributeSshKeysState.class);

    private final SetupSoftwareExperimentJob setupSoftwareExperimentJob;
    private final List<DistributeSshKeysStateSlice> slices;

    private final Experiment experiment;

    public DistributeSshKeysState(SetupSoftwareExperimentJob job) {
        super("Distribute generated SSH keys");
        this.setupSoftwareExperimentJob = job;
        this.experiment = job.getExperiment();

        slices = experiment.getKeypairs().entrySet().stream()
                .map(entry ->
                        new DistributeSshKeysStateSlice(job, entry.getKey(), entry.getValue()))
                .collect(Collectors.toList());

        assert experiment.getSliceOrNull() != null;
    }

    @Override
    public List<DistributeSshKeysStateSlice> getSlices() {
        return slices;
    }


    public static boolean hasDistributeSshKeyPair(Experiment experiment) {
        return !getDistributeSshKeyPairs(experiment).isEmpty();
    }

    @Nonnull
    private static List<? extends DistributeSshKeypair> getDistributeSshKeyPairs(Experiment experiment) {
        RequestRspecSource requestRspecSource = experiment.getSlice().getRequestRspec();
        ModelRspec model = requestRspecSource == null ?
                null :
                requestRspecSource.getModelRspec();
        assert model != null;

        if (model == null) {
            LOG.warn("ExecuteAnsibleServicesState will do nothing because there is no usable request RSpec");
            return Collections.emptyList();
        }

        return Collections.unmodifiableList(model.getDistributeSshKeypairs());
    }

    public class DistributeSshKeysStateSlice extends StateSlice {
        private final DistributeSshKeypair distributeSshKeypair;
        private final KeyPair keyPair;

        private final String location;
        //private final String folder;
        //private final String filename;

        public DistributeSshKeysStateSlice(@Nonnull Job<?> job,
                                           @Nonnull DistributeSshKeypair distributeSshKeypair,
                                           @Nonnull KeyPair keyPair) {
            super(job);
            this.distributeSshKeypair = distributeSshKeypair;
            this.keyPair = keyPair;

            String location = distributeSshKeypair.getLocation();
            if (location == null)
                location = "~/.ssh/id_rsa";

            this.location = location;
        }

        @Override
        public String getName() {
            return "Distribute generated SSH key for user '" +
                    distributeSshKeypair.getUser().getResourceName() + "'";
        }

        @Override
        public ExperimentTaskStatus statefulRun() throws JFedException, InterruptedException {
            assert experiment.getSliceOrNull() != null;
            assert experiment.getSlice().getManifestRspec() != null;
            ModelRspec manifestModel =
                    experiment.getSlice().getManifestRspec().getModelRspec();
            assert manifestModel != null;

            if (location == null) {
                return ExperimentTaskStatus.FAILED;
            }

            boolean allSuccess = true;
            boolean allFailed = true;

            for (RspecNode node : manifestModel.getNodes()) {
                final UploadKeyToNodeState uploadKeyToNodeState = new UploadKeyToNodeState(node);
                setAndRunState(uploadKeyToNodeState);

                if (uploadKeyToNodeState.getStatus() != ExperimentTaskStatus.SUCCESS)
                    allSuccess = false;
                if (uploadKeyToNodeState.getStatus() != ExperimentTaskStatus.FAILED)
                    allFailed = false;
            }
            return allSuccess ? ExperimentTaskStatus.SUCCESS :
                    allFailed ? ExperimentTaskStatus.FAILED : ExperimentTaskStatus.WARNING;
        }


        public class UploadKeyToNodeState extends State {
            private final RspecNode node;

            protected UploadKeyToNodeState(RspecNode node) {
                super("Upload generated key for '" + distributeSshKeypair.getUser().getResourceName() + "' to '" + node.getClientId() + "'");
                this.node = node;
            }

            @Nonnull
            @Override
            protected ExperimentTaskStatus executeState(Job<?> job) throws InterruptedException, JFedException {
                SSHClient ssh;
                try {
                    ssh = setupSoftwareExperimentJob.getSSHClient(node, null);
                } catch (SetupSoftwareExperimentJob.SshException e) {
                    LOG.warn("Aborting UploadKeyToNodeState.statefulRun due to SSH comnection problem", e);
                    return ExperimentTaskStatus.FAILED;
                }
                assert ssh != null;
                ExternalFileUtil externalFileUtil = new ExternalFileUtil(ssh);

                //connect to the node and deposit the key
                try {
                    char[] keyFile = KeyUtil.privateKeyToPem(keyPair.getPrivate());

                    String localFolder = externalFileUtil.getFolder(location);
                    String filename = externalFileUtil.getBasename(location);

                    try (Session session = ssh.startSession()) {
                        Session.Command command = session.exec("mkdir -p " + localFolder.replace(" ", "\\ "));
                        command.join();
                        if (command.getExitStatus() != 0) {
                            String errorMessage;
                            try (BufferedReader buffer = new BufferedReader(new InputStreamReader(command.getErrorStream()))) {
                                errorMessage = buffer.lines().collect(Collectors.joining(" "));
                            }
                            LOG.warn("Could not create directory to upload key into: {}", errorMessage);
                            updateMessage("Could not create directory to upload key into: " + errorMessage);
                        } else {
                            LOG.info("mkdir command for '{}' completed succesfully", localFolder);
                        }
                    }

                    ssh.newSCPFileTransfer().upload(new StringMemorySourceFile(filename, keyFile), localFolder);
                    LOG.debug("Succesfully uploaded private key to {} on {}", distributeSshKeypair.getLocation(), node.getClientId());

                    ssh.startSession().exec("chmod 600 " + localFolder + "/" + filename);

                    return ExperimentTaskStatus.SUCCESS;
                } catch (UserAuthException ex) {
                    LOG.error("Could not authenticate on node {} with generated key. Skipping.", node.getClientId(), ex);
                    updateMessage("Could not authenticate with generated key.");
                    return ExperimentTaskStatus.FAILED;
                } catch (IOException ex) {
                    LOG.error("Error while transferring private key {} to {}", distributeSshKeypair.getLocation(), node.getClientId(), ex);
                    updateMessage("Error while transferring private key: " + ex.getMessage());
                    return ExperimentTaskStatus.FAILED;
                } finally {
                    setupSoftwareExperimentJob.closeSSH(node);
                }
            }
        }
    }
}
